#include "job.h"
#include "tcp.h"

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <sys/socket.h>

FILE *jobsf = NULL;

/*
 * Terminate the program.
 * If err is not NULL, an error will be printed with perror.
 */
static void term(int code, char *err)
{
	if (err != NULL)
	{
		perror(err);
	}

	printf("closing file\n");

	if (jobsf != NULL)
		fclose(jobsf);

	exit(code);
}

/*
 * Handle requests.
 *
 * Input:
 *     buf: A buffer of size BUF_SIZE. Contains the request payload.
 *     sockfd: The socket to write a response to.
 *
 * Return:
 *     1 when done reading, 0 otherwise
 */
static int request_handler(int sockfd)
{
	char req[5]; // max length: 'G'<4 bytes>
	memset(req, 0, sizeof(req));

	int readlen = read(sockfd, req, sizeof(req));
	if (readlen == 0)
	{
		fprintf(stderr, "Client disconnected.\n");
		rewind(jobsf);
		return 1;
	}
	else if (readlen < 0)
	{
		fprintf(stderr, "Client read failed.\n");
		rewind(jobsf);
		return 1;
	}

	printf("Got message type: %c\n", req[0]);

	if (req[0] == 'G')
	{
		char res[sizeof(job)];
		memset(res, 0, sizeof(res));

		uint32_t count = 0;
		count |= (uint32_t)req[1];
		count |= (uint32_t)req[2] << 8;
		count |= (uint32_t)req[3] << 16;
		count |= (uint32_t)req[4] << 24;
		if (count == 0)
			count = 1;

		for (int i = 0; i < count; ++i)
		{
			job j;
			job_read(jobsf, &j);

			if (j.type == 0)
				j.type = 'Q';

			job_stringify(res, &j);
			if (write(sockfd, res, sizeof(res)) < 0)
				perror("Client write failed.");

			if (j.type == 'Q')
				break;
		}
	}
	else if (req[0] == 'T')
	{
		close(sockfd);
		printf("Client exited normally. Exiting.\n");
		term(0, NULL);
	}
	else if (req[0] == 'E')
	{
		close(sockfd);
		fprintf(stderr, "Client exited because of an error!\n");
		term(1, NULL);
		return 1;
	}
	else
	{
		fprintf(stderr, "Unknown request code: %c\n", req[0]);
	}

	return 0;
}

/*
 * SIGINT/SIGTERM handler
 */
void onterm(int signum)
{
	printf("\n");
	term(1, NULL);
}

int main(int argc, char *argv[])
{
	// SIGTERM and SIGINT handler
	struct sigaction action;
	memset(&action, 0, sizeof(action));
	action.sa_handler = onterm;
	sigaction(SIGTERM, &action, NULL);
	sigaction(SIGINT, &action, NULL);

	if (argc < 3)
	{
		printf("Usage: %s <joblist> <port>\n", argv[0]);
		return 1;
	}

	char *joblist = argv[1];
	int port = atoi(argv[2]);

	jobsf = fopen(joblist, "r");
	if (jobsf == NULL)
	{
		perror(joblist);
		return 1;
	}

	if (tcp_create_server(port, &request_handler) < 0)
		term(1, "Couldn't create server");

	fprintf(stderr, "Client disconnected before completion. Exiting.\n");
	return 1;
}
